@@ -1,10 +1,10 @@
 use strictures 2;
-package OpenAPI::Modern; # git description: v0.122-5-g3c285fe6
+package OpenAPI::Modern; # git description: v0.123-7-g3ac7d3ef
 # vim: set ts=8 sts=2 sw=2 tw=100 et :
 # ABSTRACT: Validate HTTP requests and responses against an OpenAPI v3.0, v3.1 or v3.2 document
 # KEYWORDS: validation evaluation JSON Schema OpenAPI v3.0 v3.1 v3.2 Swagger HTTP request response
 
-our $VERSION = '0.123';
+our $VERSION = '0.124';
 
 use 5.020;
 use utf8;
@@ -29,7 +29,7 @@ use Feature::Compat::Try;
 use Encode 2.89 ();
 use JSON::Schema::Modern;
 use JSON::Schema::Modern::Utilities qw(jsonp unjsonp canonical_uri E abort is_equal true false get_type);
-use OpenAPI::Modern::Utilities qw(add_vocab_and_default_schemas intersect_types coerce_primitive);
+use OpenAPI::Modern::Utilities qw(add_vocab_and_default_schemas uri_decode intersect_types coerce_primitive);
 use JSON::Schema::Modern::Document::OpenAPI;
 use MooX::TypeTiny 0.002002;
 use Types::Standard qw(InstanceOf Bool);
@@ -150,18 +150,20 @@ sub validate_request ($self, $request, $options = {}) {
           $request_parameters_processed->{$param_obj->{in}}{$fc_name} = $section;
         }
 
+        $state->{depth} += 1;
+
         my $valid =
-            $param_obj->{in} eq 'path' ? $self->_validate_path_parameter({ %$state, depth => $state->{depth}+1 }, $param_obj, $options->{path_captures})
-          : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter({ %$state, depth => $state->{depth}+1 }, $param_obj, $request->url)
-          : $param_obj->{in} eq 'header' ? $self->_validate_header_parameter({ %$state, depth => $state->{depth}+1 }, $param_obj->{name}, $param_obj, $request->headers)
-          : $param_obj->{in} eq 'cookie' ? $self->_validate_cookie_parameter({ %$state, depth => $state->{depth}+1 }, $param_obj)
-          : $param_obj->{in} eq 'querystring' ? $self->_validate_querystring_parameter({ %$state, depth => $state->{depth}+1 }, $param_obj, $request->url)
+            $param_obj->{in} eq 'path' ? $self->_validate_path_parameter({ %$state, data_path => '/request/uri/path' }, $param_obj, $options->{path_captures})
+          : $param_obj->{in} eq 'query' ? $self->_validate_query_parameter({ %$state, data_path => '/request/uri/query' }, $param_obj, $request->url)
+          : $param_obj->{in} eq 'header' ? $self->_validate_header_parameter({ %$state, data_path => '/request/header' }, $param_obj->{name}, $param_obj, $request->headers)
+          : $param_obj->{in} eq 'cookie' ? $self->_validate_cookie_parameter({ %$state, data_path => '/request/header/Cookie' }, $param_obj)
+          : $param_obj->{in} eq 'querystring' ? $self->_validate_querystring_parameter({ %$state, data_path => '/request/uri/query' }, $param_obj, $request->url)
           : abort($state, 'unrecognized "in" value "%s"', $param_obj->{in});
       }
     }
 
-    # v3.2.0 §4.8.2, "Path Templating": "Each template expression in the path MUST correspond to a path
-    # parameter that is included in the Path Item itself and/or in each of the Path Item’s
+    # v3.2.0 §4.8.2, "Path Templating": "Each template expression in the path MUST correspond to a
+    # path parameter that is included in the Path Item itself and/or in each of the Path Item’s
     # Operations."
     # We could validate this at document parse time, except the path-item can also be reached via a
     # $ref and the referencing path could be from another document and is therefore unknowable until
@@ -312,7 +314,7 @@ sub validate_response ($self, $response, $options = {}) {
         $header_obj = $self->_resolve_ref('header', $ref, $state);
       }
 
-      ()= $self->_validate_header_parameter({ %$state, depth => $state->{depth}+1 },
+      ()= $self->_validate_header_parameter({ %$state, data_path => '/response/header', depth => $state->{depth}+1 },
         $header_name, $header_obj, $response->headers);
     }
 
@@ -649,7 +651,8 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
 
   # extract all capture values from path template variables: ($1 .. $n)
   # perldoc perlvar, @-: $n coincides with "substr $_, $-[n], $+[n] - $-[n]" if "$-[n]" is defined
-  my @path_capture_values = map _uri_decode(substr($uri, $-[$_], $+[$_]-$-[$_])), 1 .. $#-;
+  # We do not url-decode here, because we may need to parse the string for style delimiters first
+  my @path_capture_values = map substr($uri, $-[$_], $+[$_]-$-[$_]), 1 .. $#-;
 
   # we set aside $state for potential restoration because we might still encounter issues later on
   # that require us to keep iterating for another URI match
@@ -722,7 +725,7 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
     my $host_variable_count = ()= ($normalized_server_url->host//'') =~ /\x00/g;
     @server_capture_values = (
       (map +(/^xn--(.+)\z/ ? punycode_decode($1) : $_), @server_capture_values[0 .. $host_variable_count-1]),
-      (map _uri_decode($_), @server_capture_values[$host_variable_count .. $#server_capture_values]));
+      (map uri_decode($_), @server_capture_values[$host_variable_count .. $#server_capture_values]));
 
     # we have a match, so preserve our new $state values created via _resolve_ref
     %$state = %$local_state;
@@ -768,43 +771,37 @@ sub _match_uri ($self, $method, $uri, $path_template, $state) {
 }
 
 sub _validate_path_parameter ($self, $state, $param_obj, $path_captures) {
-  $state->{data_path} .= '/uri/path';
-
   # 'required' is always true for path parameters
+  # v3.2.0 §4.12.2.1: "If "in" is "path", the name field MUST correspond to a single template
+  # expression occurring within the path field in the Paths Object."
   return E({ %$state, keyword => 'required' }, 'missing path parameter: %s', $param_obj->{name})
     if not exists $path_captures->{$param_obj->{name}};
 
-  $state->{data_path} = jsonp($state->{data_path}, $param_obj->{name});
+  my $data = $path_captures->{$param_obj->{name}}.'';
 
-  my $data = $path_captures->{$param_obj->{name}};
-  $data .= '';
-
-  return $self->_validate_parameter_content({ %$state, depth => $state->{depth}+1 }, $param_obj, \$data)
-    if exists $param_obj->{content};
-
-  return E({ %$state, keyword => 'style' }, 'only style: simple is supported in path parameters')
-    if ($param_obj->{style}//'simple') ne 'simple';
-
-  my @types = $self->_type_in_schema($param_obj->{schema}, { %$state, keyword_path => $state->{keyword_path}.'/schema' });
-
-  if (any { $_ eq 'string' || $_ eq 'number' || $_ eq 'boolean' || $_ eq 'null' } @types) {
-    return E($state, 'cannot deserialize to %s type%s', 'requested', @types > 1 ? 's' : '')
-      if not coerce_primitive(\$data, \@types);
+  if (exists $param_obj->{content}) {
+    $data = uri_decode($data);
+    return $self->_validate_parameter_content({ %$state, depth => $state->{depth}+1,
+      data_path => jsonp($state->{data_path}, $param_obj->{name}) }, $param_obj, \$data);
   }
-  elsif (any { $_ eq 'array' || $_ eq 'object'} @types) {
-    return E($state, 'deserializing path parameters to arrays or objects is not currently supported');
-  }
-  else {
-    return E($state, 'cannot deserialize to %s type%s', !@types ? 'any' : 'requested', @types > 1 ? 's' : '');
+
+  $data = $self->_deserialize_style($data, my $s = { %$state, errors => [] },
+    style => $param_obj->{style}//'simple',
+    explode => $param_obj->{explode}//false,
+    $param_obj->%{qw(name schema)},
+  );
+
+  if ($s->{errors}->@*) {
+    push $state->{errors}->@*, $s->{errors}->@*;
+    return;
   }
 
   $self->_evaluate_subschema(\$data, $param_obj->{schema},
-    { %$state, keyword_path => $state->{keyword_path}.'/schema', depth => $state->{depth}+1 });
+    { %$state, data_path => jsonp($state->{data_path}, $param_obj->{name}),
+      keyword_path => $state->{keyword_path}.'/schema', depth => $state->{depth}+1 });
 }
 
 sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
-  $state->{data_path} .= '/uri/query';
-
   # parse the query parameters out of uri
   my $query_params = +{ $uri->query->pairs->@* };
 
@@ -858,8 +855,6 @@ sub _validate_query_parameter ($self, $state, $param_obj, $uri) {
 sub _validate_header_parameter ($self, $state, $header_name, $header_obj, $headers) {
   return 1 if grep fc $header_name eq fc $_, qw(Accept Content-Type Authorization);
 
-  $state->{data_path} .= '/header';
-
   if (not $headers->every_header($header_name)->@*) {
     return E({ %$state, keyword => 'required' }, 'missing header: %s', $header_name)
       if $header_obj->{required};
@@ -921,14 +916,10 @@ sub _validate_header_parameter ($self, $state, $header_name, $header_obj, $heade
 }
 
 sub _validate_cookie_parameter ($self, $state, $param_obj, @args) {
-  $state->{data_path} = jsonp($state->{data_path}, qw(headers Cookie), $param_obj->{name});
-
   return E($state, 'cookie parameters not yet supported');
 }
 
 sub _validate_querystring_parameter ($self, $state, $param_obj, $uri) {
-  $state->{data_path} = '/request/uri/query';
-
   # note: if something has caused the Mojo::Parameters object to be normalized (e.g. calling
   # 'pairs'), the raw string value is lost
   return E({ %$state, keyword => 'required' }, 'missing querystring')
@@ -941,6 +932,167 @@ sub _validate_querystring_parameter ($self, $state, $param_obj, $uri) {
   $self->_validate_parameter_content({ %$state, depth => $state->{depth}+1 }, $param_obj, \$content)
 }
 
+# data comes in as a string
+# when parsing fails, $state->{errors} is populated;
+# otherwise, the return value is the fully deserialized data, parsed into the correct type(s)
+# (which may be undef = null; note the difference from () which indicates an error)
+# This method is not appropriate for header parameters, which should never be percent-decoded.
+# %opt is (:$style, :$explode, :$name, :$schema)
+sub _deserialize_style ($self, $data, $state, %opt) {
+  # numbers and builtin bools can be treated as strings, but reject undef and references
+  croak 'only strings can be deserialized' if not defined $data or ref $data;
+
+  my ($style, $explode, $name, $schema) = @opt{qw(style explode name schema)};
+  my @types = $self->_type_in_schema($schema, { %$state, keyword_path => $state->{keyword_path}.'/schema' });
+
+  if ($style eq 'simple' or $style eq 'matrix' or $style eq 'label') {
+    my $state = +{ %$state, data_path => jsonp($state->{data_path}, $name) };
+
+    # RFC6570 §3.2.1: "A variable that is undefined (Section 2.3) has no value and is ignored by the
+    # expansion process. If all of the variables in an expression are undefined, then the
+    # expression's expansion is the empty string."
+    if ($data eq '') {
+      if (any { $_ eq 'null' } @types) {
+        return undef;
+      }
+      elsif (any { $_ eq 'array' } @types) {
+        # RFC6570 §2.3-6: "A variable defined as a list value is considered undefined if the list
+        # contains zero members."
+        return [];
+      }
+      elsif (any { $_ eq 'object' } @types) {
+        # RFC6570 §2.3-6: "A variable defined as an associative array of (name, value) pairs is
+        # considered undefined if the array contains zero members or if all member names in the
+        # array are associated with undefined values."
+        return {};
+      }
+    }
+
+    my $prefix =
+        $style eq 'simple' ? ''
+      : $style eq 'label' ? '\.'
+      : $style eq 'matrix' ? ';'
+      : die;
+
+    return E({ %$state, keyword => 'style' },
+        'data does not match indicated style "%s" (invalid prefix)', $style)
+      if length $prefix and $data !~ s/^$prefix//;
+
+    my $delimiter =
+        !$explode ? ','
+      : $style eq 'simple' ? ','
+      : $style eq 'matrix' ? ';'
+      : $style eq 'label' ? '\.'
+      : die;
+
+    my @errors;
+
+    # if all types are acceptable, fall through to returning string immediately
+
+    # this block is for:
+    # style=simple, explode=true, object
+    # style=label, explode=true, object
+    # style=matrix, explode=true, object
+    # style=matrix, explode=true, array
+    # but NOT: explode=false (any style), or style=simple/label, explode=true for arrays.
+    if (@types != 6 and $explode
+        and ((($style eq 'simple' or $style eq 'label') and any { $_ eq 'object' } @types)
+          or ($style eq 'matrix' and any { $_ eq 'object' || $_ eq 'array' } @types))) {
+
+      my @values = split($delimiter, $data, -1);
+
+      my $type = (any { $_ eq 'object' } @types) ? 'object' : 'array';
+      my $idx = -1;
+
+      # we only make one attempt, even if both types are requested, because any errors when
+      # deserializing to an object will also occur for array
+      # RFC6570 §3.2.1-7: for arrays with style=matrix and explode=true, expand as an object,
+      # where the key names are the parameter name
+
+      my @keys_and_values = map uri_decode($_),
+        map {
+          ++$idx;
+          # RFC6570 §3.2.1-6: empty value does not use '='
+          my ($key, $val) = split(/=/, $_, -1);
+          ()= E({ %$state, keyword => 'style', errors => \@errors },
+              'data does not match indicated style "%s" for %s (invalid separator at %s)',
+              $style, $type, $type eq 'object' ? 'key "'.$key.'"' : 'index '.$idx)
+            if defined $val and not length $val;
+          ($key, $val//'');
+        }
+        @values;
+
+      my $data = $type eq 'object' ? +{ @keys_and_values }
+        : [ map {
+            ()= E({ %$state, keyword => 'style', errors => \@errors },
+                'data does not match indicated style "matrix" for array (invalid element name%s)',
+                defined $_->[0] ? ' at "'.$_->[0].'"' : '')
+              if not defined $_->[0] or $_->[0] ne $name;
+            $_->[1];
+          }
+          pairs @keys_and_values
+        ];
+
+      if (not @errors) {
+        $self->_coerce_object_elements($data, $schema, { %$state, keyword_path => $state->{keyword_path}.'/schema' }) if $type eq 'object';
+        $self->_coerce_array_elements($data, $schema, { %$state, keyword_path => $state->{keyword_path}.'/schema' }) if $type eq 'array';
+        return $data;
+      }
+    }
+
+    # process matrix prefix for primitives, and for array and object when explode=false
+    if ($style eq 'matrix'
+        and (not $explode or any { $_ eq 'string' || $_ eq 'number' || $_ eq 'boolean' || $_ eq 'null' } @types)
+        # '=' is included after the name iff the variable's value is not empty; name is still encoded
+        and ($data !~ s/^([^{}=]+)(?:=(?=.)|$)// or uri_decode($1) ne $name)) {
+      ()= E({ %$state, keyword => 'style', errors => \@errors },
+        'data does not match indicated style "matrix" (invalid prefix)');
+      push $state->{errors}->@*, @errors;
+      return;
+    }
+
+    # this block is for:
+    # all styles, explode=false, objects
+    # style=simple or label, explode=false, array
+    # style=simple or label, explode=true, array
+    # style=matrix, explode=false, array
+
+    if (@types != 6 and any { $_ eq 'array' || $_ eq 'object' } @types) {
+      my @values = map uri_decode($_), split($delimiter, $data, -1);
+
+      if (not $explode and any { $_ eq 'object' } @types) {
+        if (not @values % 2) {
+          $data = +{ @values };
+          $self->_coerce_object_elements($data, $schema, { %$state, keyword_path => $state->{keyword_path}.'/schema' });
+          return $data;
+        }
+        # fall through to primitive
+      }
+
+      if (any { $_ eq 'array' } @types
+          and (any { $style eq $_ } qw(simple label) or ($style eq 'matrix' and not $explode))) {
+        $data = \@values;
+        $self->_coerce_array_elements($data, $schema, { %$state, keyword_path => $state->{keyword_path}.'/schema' });
+        return $data;
+      }
+    }
+
+    $data = uri_decode($data);
+    return $data if coerce_primitive(\$data, \@types);
+
+    if (@errors) {
+      push $state->{errors}->@*, @errors;
+      return;
+    }
+  }
+  else {
+    die 'unsupported style ', $style;
+  }
+
+  return E({ %$state, data_path => jsonp($state->{data_path}, $name) },
+    'cannot deserialize to %s type%s', !@types ? 'any' : 'requested', @types > 1 ? 's' : '');
+}
+
 sub _validate_parameter_content ($self, $state, $param_obj, $content_ref) {
   my ($media_type) = keys $param_obj->{content}->%*;  # there can only be one key
 
@@ -1460,11 +1612,6 @@ sub _convert_response ($response) {
   return $res;
 }
 
-# url-percent-decode and UTF-8-decode a string
-sub _uri_decode ($str) {
-  Encode::decode('UTF-8', url_unescape($str), Encode::DIE_ON_ERR);
-}
-
 # callback hook for Sereal::Encoder
 sub FREEZE ($self, $serializer) { +{ %$self } }
 
@@ -1494,7 +1641,7 @@ OpenAPI::Modern - Validate HTTP requests and responses against an OpenAPI v3.0,
 
 =head1 VERSION
 
-version 0.123
+version 0.124
 
 =head1 SYNOPSIS
 
@@ -1783,7 +1930,7 @@ C<operation_id>: a string corresponding to the L<operationId|https://learn.opena
 
 =item *
 
-C<path_captures>: a hashref mapping placeholders in the path template to their actual values in the request URI; these will be url-unescaped
+C<path_captures>: a hashref mapping placeholders in the path template to their literal values in the request URI (not percent-decoded; not recommended to be provided for use in path matching due to inconsistencies with url-escaping, and this capability may be removed in the future)
 
 =back
 
@@ -1805,7 +1952,7 @@ In addition, these values are also populated in the options hash (when available
 
 =item *
 
-C<uri_captures>: a hashref mapping placeholders in the entire uri template (server url plus path template) to their actual values in the request URI; these will be url-unescaped (for values coming from the path) and punycode-decoded (for values coming from the host).
+C<uri_captures>: a hashref mapping placeholders in the entire uri template (server url plus path template) to their actual values in the request URI; for values coming from the host, these will be puny-decoded; for values coming from the path via a server url these will be url-unescaped, but path values coming from the path template are B<NOT> url-unescaped.
 
 =item *
 
@@ -1983,10 +2130,6 @@ Only certain permutations of OpenAPI documents are supported at this time:
 
 =item *
 
-for path parameters, only C<style: simple> and C<explode: false> is supported
-
-=item *
-
 for query parameters, only C<style: form> and C<explode: true> is supported, only the first value of each parameter name is considered, and C<allowEmptyValue> and C<allowReserved> are not checked
 
 =item *
